from django.db import models
from django.utils import translation
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.http import HttpRequest

from wagtail.log_actions import log
from wagtail.blocks import (
    StreamValue,
)
from wagtail.models import (
    RevisionMixin,
)

from .base import (
    BlockFieldReplacementAdapter,
    AdapterError,
    Keyword,
    VARIABLES,
)
from ..forms import (
    blocks as block_forms,
)
from .. import utils
from wagtail.admin.admin_url_finder import (
    AdminURLFinder,
)
from ..toolbar import (
    FeditAdapterComponent,
    FeditAdapterAdminLinkButton,
)



class BlockAdapter(BlockFieldReplacementAdapter):
    """
    An adapter for editing Wagtail blocks.
    This will render the block and replace it on the frontend
    on successful form submission.
    """
    identifier = "block"
    usage_description = "This adapter is used to edit a block of a streamfield."
    keywords = BlockFieldReplacementAdapter.keywords + (
        Keyword("block",
            help_text="the block instance to edit. This can be a regular block instance or a BoundBlock.",
            type_hint="blocks.Block"
        ),
        Keyword("block_id",
            optional=True,
            help_text="the ID of the block to edit, required if block is not a BoundBlock.",
            type_hint="str"
        ),
        Keyword("admin",
            absolute=True,
            help_text="if passed; the adapter will add a quick- link to the Wagtail Admin for this block."
        ),
    )

    def __init__(self, object: models.Model, field_name: str, request: HttpRequest, **kwargs):
        super().__init__(object, field_name, request, **kwargs)

        self.block = self.kwargs.pop("block", None)
        if self.block:
            if not hasattr(self.block, "id") and not self.kwargs["block_id"]:
                raise AdapterError("Invalid block type, block must have an `id` attribute or provide a `block_id`")
            
            if hasattr(self.block, "id"):
                self.kwargs["block_id"] = self.block.id

        else:
            block_id = self.kwargs["block_id"]
            if block_id is None:
                raise AdapterError("Block ID is required")
            
            self.streamfield: StreamValue = getattr(self.object, self.field_name)
            result = utils.find_block(block_id, self.streamfield)
            if not result:
                raise AdapterError("Block not found; did you provide the correct block ID?")
            
            self.block, _ = result

    @cached_property
    def tooltip(self) -> str:
        return self.get_header_title()

    def get_admin_url(self) -> str:
        finder = AdminURLFinder(self.request.user)
        url = finder.get_edit_url(self.object)
        hash = f"#block-{self.kwargs['block_id']}-section"
        return f"{url}{hash}"

    def get_toolbar_buttons(self) -> list[FeditAdapterComponent]:
        buttons = super().get_toolbar_buttons()
        if not self.kwargs["admin"]:
            return buttons
        
        buttons.append(FeditAdapterAdminLinkButton(
            self.request, self,
        ))
        return buttons

    def get_header_title(self):

        model_string = getattr(self.object, "get_admin_display_title", None)
        if model_string:
            model_string = model_string()
        else:
            model_string = getattr(self.object, "title", str(self.object))

        return _("Edit block %(block_label)s for %(model_name)s '%(model_string)s'") % {
            "block_label": self.block.block.label,
            "model_name": self.model._meta.verbose_name,
            "model_string": model_string,
        }
    
    def get_element_id(self) -> str:
        return f"block-{self.kwargs['block_id']}-section"
    
    def get_form_attrs(self) -> dict:

        size = getattr(self.block.block.meta, VARIABLES.PY_SIZE_VAR, None)
        if size:
            return super().get_form_attrs() | {
                VARIABLES.FORM_SIZE_VAR: size,
            }
        
        return {}
    
    def check_permissions(self):
        return super().check_permissions() and getattr(
            self.block.block.meta, "feditable", True
        )

    def get_form(self):

        self.form_class = block_forms.get_block_form_class(self.block.block)

        if self.request.method == "POST":
            form = self.form_class(self.request.POST, block=self.block, parent_instance=self.object, request=self.request)
        else:
            form = self.form_class(block=self.block, parent_instance=self.object, request=self.request)
        return form
    
    def form_valid(self, form: block_forms.BlockEditForm):
        self.block = form.save()

        extra_log_kwargs = {}
        if isinstance(self.object, RevisionMixin):
            extra_log_kwargs["revision"] = self.object.latest_revision

        with translation.override(None):
            log(
                instance=self.object,
                action="wagtail_fedit.edit_block",
                user=self.request.user,
                title=self.get_header_title(),
                data={
                    "block_id": self.kwargs["block_id"],
                    "field_name": self.field_name,
                    "model_id": self.object.pk,
                    "model_name": self.object._meta.model_name,
                    "app_label": self.object._meta.app_label,
                    "verbose_field_name": str(self.meta_field.verbose_name),
                    "block_label": str(self.block.block.label),
                },
                content_changed=True,
                **extra_log_kwargs,
            )

    @classmethod
    def render_from_kwargs(cls, context, **kwargs):
        if "block" not in kwargs:
            raise AdapterError("Block is required")
        
        block = kwargs.pop("block")
        if not hasattr(block, "render"):
            raise AdapterError("Invalid block type, missing render method")
        
        return block.render(context)

    def render_content(self, parent_context: dict = None) -> str:
        parent_context = parent_context or {}
        if hasattr(parent_context, "flatten"):
            parent_context = parent_context.flatten()

        return self.block.render(parent_context)
        


from .adapters import (
    BaseAdapterView,
    EditAdapterView,
    AdapterRefetchView,
)
from .editable import (
    FEditableView,
    PublishView,
    UnpublishView,
    SubmitView,
    CancelView,
)


from typing import Any
from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext_lazy as _
from django.utils import translation
from django.urls import reverse
from django.shortcuts import redirect
from django.views.generic import TemplateView
from django.http import (
    HttpRequest,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponse,
)
from wagtail.admin import messages
from wagtail.log_actions import (
    registry,
)
from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.actions.publish_page_revision import PublishPageRevisionAction
from wagtail.actions.publish_revision import PublishRevisionAction
from wagtail.actions.unpublish_page import UnpublishPageAction
from wagtail.actions.unpublish import UnpublishAction
from wagtail.models import (
    RevisionMixin,
    PreviewableMixin,
    DraftStateMixin,
    WorkflowMixin,
    WorkflowState,
    PageLogEntry,
    ModelLogEntry,
    Page,
)
from ..utils import (
    FEDIT_PREVIEW_VAR,
    USERBAR_MODEL_VAR,
    LOG_ACTION_TEMPLATES_AVAILABLE,
    FeditPermissionCheck,
    with_userbar_model,
    # user_can_publish,
    # user_can_unpublish,
    # user_can_submit_for_moderation,
    # lock_info,
)
from ..errors import (
    NO_PERMISSION_VIEW,
    NO_PERMISSION_ACTION,
    NO_WORKFLOW_STATE,
    OBJECT_LOCKED,
    OBJECT_NOT_LIVE,
    ACTION_MISSING,
    INVALID_ACTION,
    MISSING_REQUIRED_SUPERCLASSES,
    NO_UNPUBLISHED_CHANGES,
)
from .mixins import (
    ObjectViewMixin,
    LockViewMixin,
    LocaleMixin,
)


MAX_LOG_ENTRIES_DISPLAYED = 5


def get_unpublish_action(object):
    if isinstance(object, Page):
        return UnpublishPageAction
    return UnpublishAction


def get_publish_action(object):
    if isinstance(object, Page):
        return PublishPageRevisionAction
    return PublishRevisionAction



class BaseFeditView(LocaleMixin, ObjectViewMixin, FeditPermissionCheck, TemplateView):
    def dispatch(self, request: HttpRequest, object_id: Any, app_label: str, model_name: str) -> HttpResponse:
        if self.error_response:
            return self.error_response

        if not self.has_perms(request, self.object):
            return HttpResponseForbidden(NO_PERMISSION_VIEW)

        if issubclass(self.model, RevisionMixin) and self.object.latest_revision_id:
            instance: RevisionMixin  = self.object
            revision: RevisionMixin = instance.latest_revision
            self.object = revision.as_object()
            self.is_preview = True
        else:
            self.is_preview = False

        try:
            self.checks(request, self.object)
        except ValueError as e:
            return HttpResponseForbidden(str(e))

        return super().dispatch(request, object_id, app_label, model_name)
    
    def checks(self, request: HttpRequest, object: Any) -> None:
        pass

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        return super().get_context_data(**kwargs) | {
            "object": self.object,
            "model": self.model,
        }


class FEditableView(BaseFeditView):

    def checks(self, request: HttpRequest, object: Any) -> None:
        super().checks(request, object)
        if not isinstance(self.object, PreviewableMixin):
            raise ValueError(MISSING_REQUIRED_SUPERCLASSES.format(self.model))

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        # # Check if lock applies to this user
        # if not self.locked_for_user:
        self.request = with_userbar_model(self.request, self.object)

        object: PreviewableMixin = self.object
        return object.make_preview_request(original_request=self.request, extra_request_attrs={
            FEDIT_PREVIEW_VAR: True,
            USERBAR_MODEL_VAR: self.object,
        })
    

class BaseActionView(LockViewMixin, BaseFeditView):
    template_name         = "wagtail_fedit/editor/action_confirm.html"
    required_superclasses = [DraftStateMixin]
    action_icon            = None
    action_text            = None
    title_format           = None
    action_help_text_title = None
    action_help_text       = None

    def get_action(self) -> str:
        return self.action_text
    
    def get_action_value(self) -> str:
        return f"{self.__class__.__name__.lower()}"

    def get_action_title(self) -> str:
        return self.title_format.format(self.object)
    
    def get_action_help_text_title(self) -> str:
        return self.action_help_text_title
    
    def get_action_help_text(self) -> str:
        return self.action_help_text
    
    def get_action_icon(self) -> str:
        return self.action_icon
    
    def setup(self, request: HttpRequest, object_id: Any, app_label: str, model_name: str) -> HttpResponse:
        super().setup(request, object_id, app_label, model_name)
        self.policy = self.object.permissions_for_user(request.user)

        if not isinstance(self.object, tuple(self.required_superclasses)):
            self.error_response = HttpResponseBadRequest(
                MISSING_REQUIRED_SUPERCLASSES.format(
                    self.model.__name__, ", ".join(
                        [cls.__name__ for cls in self.required_superclasses],
                    )
                )
            )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        return super().get_context_data(**kwargs) | {
            "action": self.get_action_value(),  # e.g. "publishview"
            "action_text": self.get_action(),
            "action_title": self.get_action_title(),
            "action_help_text_title": self.get_action_help_text_title(),
            "action_help_text": self.get_action_help_text(),
            "action_icon": self.get_action_icon(),
            "cancel_url": reverse(
                "wagtail_fedit:editable",
                args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
            ),
        }

    def redirect_to_success_url(self, request: HttpRequest) -> HttpResponse:
        if hasattr(self.object, "get_url"):
            return redirect(self.object.get_url(request))
        
        elif hasattr(self.object, "get_absolute_url"):
            return redirect(self.object.get_absolute_url())
        
        return redirect(reverse(
            "wagtail_fedit:editable",
            args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
        ))

    def redirect_to_failsafe_url(self, request: HttpRequest) -> HttpResponse:
        try:
            finder = AdminURLFinder(request.user)
            edit_url = finder.get_edit_url(self.object)
            return redirect(edit_url)
        except Exception:
            return redirect(reverse(
                "wagtail_fedit:editable",
                args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
            ))
        
    def check_policy(self, request: HttpRequest, policy: FeditPermissionCheck) -> None:
        pass

    def action(self, request: HttpRequest) -> HttpResponse:
        raise NotImplementedError("Subclasses must implement this method")
    
    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        
        if getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True) and issubclass(self.model, WorkflowMixin):
            # Retrieve current workflow state if set, default to last workflow state
            self.workflow_state = (
                self.object.current_workflow_state
                or self.object.workflow_states.order_by("created_at").last()
            )
        else:
            self.workflow_state = None

        try:
            self.check_policy(request, self.policy)
        except ValueError as e:
            messages.error(request, str(e))
            return self.get(request, *args, **kwargs)

        # Check if lock applies to this user
        if self.locked_for_user:
            messages.error(request, OBJECT_LOCKED)
            return self.redirect_to_failsafe_url(request)
        
        if "action" not in request.POST:
            messages.error(request, ACTION_MISSING)
            return redirect(request.path)
        
        if request.POST["action"] != self.get_action_value():
            messages.error(request, INVALID_ACTION.format(request.POST["action"]))
            return redirect(request.path)
        
        return self.action(request)


class PublishView(BaseActionView):
    template_name = "wagtail_fedit/editor/action_publish_confirm.html"
    required_superclasses = [DraftStateMixin, RevisionMixin]
    action_text = _("Publish")
    action_icon = "fedit-eye-open"

    def get_action_title(self):
        return _("Publishing {} \"{}\"").format(self.object._meta.verbose_name, self.object)

    def get_action_help_text_title(self):
        return _("About publishing")

    def get_action_help_text(self):
        s = [
            _("Publishing this object will make it visible to users on the site."),
            _("That means that any changes you make will be immediately visible to everyone."),
        ]

        if self.policy.can_unpublish():
            s.append(_("You can always choose to unpublish it."))

        return s
    
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)

        log_entry_count = 0
        log_entry_model = registry.get_log_model_for_model(self.object.__class__)
        if issubclass(log_entry_model, PageLogEntry):
            log_entries = log_entry_model.objects\
                .filter(page=self.object)\
                .order_by("-timestamp")
            
            context["view_more_url"] = reverse(
                "wagtailadmin_pages:history",
                args=[self.object.pk],
            )
                
        elif issubclass(log_entry_model, ModelLogEntry):
            log_entries = log_entry_model.objects\
                .filter(object_id=self.object.pk)\
                .filter(content_type=ContentType.objects.get_for_model(self.object))\
                .order_by("-timestamp")
                
        else:
            log_entries = None

        has_existing_published = log_entries.filter(action="wagtail.publish").exists()
        if log_entries and has_existing_published:
            log_entries = log_entries.filter(
                timestamp__gt=models.Subquery(
                    log_entries.filter(action="wagtail.publish")\
                               .values("timestamp")\
                               .order_by("-timestamp")[:1]
                )
            )

            log_entries = log_entries.select_related(
                "revision", "user", "user__wagtail_userprofile",
            )

            # if not self.request.user.is_superuser or\
            #    not self.request.user.is_staff:
            #     log_entries = log_entries.filter(user=self.request.user)

            if isinstance(self.object, Page):
                log_entries = log_entries.select_related("page")
            else:
                log_entries = log_entries.select_related("content_type")

            log_entry_count = log_entries.count()
            log_entries = log_entries[:MAX_LOG_ENTRIES_DISPLAYED]


        context.update({
            "log_entries": log_entries,
            "has_more_entries": log_entry_count > MAX_LOG_ENTRIES_DISPLAYED,
            "log_entry_count": log_entry_count,
            "last_published_at": self.object.last_published_at,
            "is_page": isinstance(self.object, Page),
            
            "LOG_ACTION_TEMPLATES_AVAILABLE": LOG_ACTION_TEMPLATES_AVAILABLE,
        })

        return context

    def check_policy(self, request: HttpRequest, policy: FeditPermissionCheck) -> None:
        if not policy.can_publish():
            raise ValueError(str(NO_PERMISSION_ACTION.format(
                _("publish")
            )))
        
        if self.locked_for_user:
            raise ValueError(str(OBJECT_LOCKED))
        
        if not self.object.has_unpublished_changes:
            raise ValueError(str(NO_UNPUBLISHED_CHANGES))

    def action(self, request: HttpRequest) -> HttpResponse:
        latest_revision = self.object.latest_revision

        if not latest_revision:
            latest_revision = self.object.save_revision(
                user=request.user,
            )

        action = get_publish_action(self.object)(
            revision=latest_revision,
            user=request.user,
        )

        action.execute()

        if latest_revision:
            self.object = latest_revision.as_object()

        return self.redirect_to_success_url(request)


class UnpublishView(BaseActionView):
    required_superclasses = [DraftStateMixin]
    action_icon = "fedit-eye-closed"
    action_text = _("Unpublish")
    action_help_text_title = _("About unpublishing")
    action_help_text = [
        _("Unpublishing this object will make it invisible to users on the site."),
        _("That means that it will no longer be visible to anyone."),
        _("You can always choose to publish it again."),
    ]

    def get_action_title(self):
        return _("Unpublishing {} \"{}\"").format(self.object._meta.verbose_name, self.object)
    
    def check_policy(self, request: HttpRequest, policy: FeditPermissionCheck) -> None:
        if not policy.can_unpublish():
            raise ValueError(str(NO_PERMISSION_ACTION.format(
                _("unpublish")
            )))
        
        if self.locked_for_user:
            raise ValueError(str(OBJECT_LOCKED))
        
        if not self.object.live:
            raise ValueError(str(OBJECT_NOT_LIVE))
 
    def action(self, request: HttpRequest) -> HttpResponse:
        if not self.object.live:
            messages.error(request, OBJECT_NOT_LIVE)
            return self.get(request)
        
        action = get_unpublish_action(self.object)(
            self.object,
            user=request.user,
        )

        action.execute()

        return self.redirect_to_failsafe_url(request)


class SubmitView(BaseActionView):
    required_superclasses = [DraftStateMixin, WorkflowMixin, RevisionMixin]
    action_icon = "fedit-check-list"
    action_text = _("Submit for moderation")
    action_help_text_title = _("About submitting for moderation")
    action_help_text = [
        _("Submitting this object for moderation will make it invisible to users on the site."),
        _("That means that it will no longer be visible to anyone."),
        _("You can always choose to publish it again."),
    ]

    def get_action_title(self):
        return _("Submitting {} \"{}\" for moderation").format(self.object._meta.verbose_name, self.object)

    def check_policy(self, request: HttpRequest, policy: FeditPermissionCheck) -> None:
        if not policy.can_submit_for_moderation():
            raise ValueError(str(NO_PERMISSION_ACTION.format(
                _("submit for moderation")
            )))
        
        if not self.object.has_unpublished_changes:
            raise ValueError(str(NO_UNPUBLISHED_CHANGES))
   
    def action(self, request: HttpRequest) -> HttpResponse:
        
        latest_revision = getattr(self.object, "latest_revision", None)
        if not latest_revision:
            latest_revision = self.object.save_revision(
                user=request.user,
            )

        if (
            self.workflow_state
            and self.workflow_state.status == WorkflowState.STATUS_NEEDS_CHANGES
        ):
            # If the workflow was in the needs changes state, resume the existing workflow on submission
            self.workflow_state.resume(self.request.user)
        else:
            # Otherwise start a new workflow
            workflow = self.object.get_workflow()
            workflow.start(self.object, self.request.user)

        return self.redirect_to_success_url(request)


class CancelView(BaseActionView):
    required_superclasses = [DraftStateMixin, WorkflowMixin]
    action_icon = "fedit-stop-sign"
    action_text = _("Cancel Workflow")
    action_help_text_title = _("About cancelling workflows")
    action_help_text = [
        _("Cancelling the workflow for this object will remove it from the moderation process."),
        _("That means that it will no longer be visible to anyone."),
        _("You can always choose to submit it for moderation again."),
    ]

    def get_action_title(self):
        return _("Cancelling workflow for {} \"{}\"").format(self.object._meta.verbose_name, self.object)

    def check_policy(self, request: HttpRequest, policy: FeditPermissionCheck) -> None:
        if not self.workflow_state:
            raise ValueError(str(NO_WORKFLOW_STATE))
        
        if not self.workflow_state.user_can_cancel(request.user):
            raise ValueError(str(NO_PERMISSION_ACTION.format(
                _("cancel the workflow")
            )))

    def action(self, request: HttpRequest) -> HttpResponse:
        if self.workflow_state:
            self.workflow_state.cancel(user=self.request.user)
            return self.redirect_to_success_url(request)
        
        return self.get(request)



from django.db.models.base import Model as Model
from django.http import HttpRequest
from django.template import (
    Context, Template,
    TemplateSyntaxError,
)
from wagtail_fedit.adapters import (
    Keyword,
    BaseAdapter,
    adapter_registry,
    BlockAdapter,
    FieldAdapter,
    ModelAdapter,
)
from wagtail_fedit.adapters.base import (
    BlockFieldReplacementAdapter,
)
from wagtail_fedit.utils import (
    FEDIT_PREVIEW_VAR,
    FIELD_TEMPLATE_VAR,
    INSTANCE_TEMPLATE_VAR,
    base_adapter_context,
    find_block,
)
from wagtail_fedit.templatetags.fedit import (
    wrap_adapter,
)
from .base import (
    BaseFEditTest,
)

import json

adapters = {}

class TestAdapter(BlockFieldReplacementAdapter):
    identifier = "test"
    keywords = (
        Keyword("test", help_text="A test keyword argument", type_hint="str"),
    )
    js_constructor = "wagtail_fedit.ThisDoesntGetUsedAnyways"

    def __init__(self, object: Model, field_name: str, request: HttpRequest, **kwargs):
        super().__init__(object, field_name, request, **kwargs)
        adapters[self.kwargs["id"]] = self

    def get_element_id(self) -> str:
        return f"test-{self.kwargs['id']}"

    def render_content(self, parent_context: dict = None) -> str:
        if "test" not in parent_context:
            return f"TestAdapter: {self.field_value} ({parent_context['id']})"
        return f"TestAdapter: {self.field_value} ({parent_context['test']}) ({parent_context['id']})"

class TestContextAdapter(TestAdapter):
    identifier = "test_context"

    def render_content(self, parent_context: dict = None) -> str:
        return parent_context["testing"]

class TestAbsoluteTokensAdapter(TestAdapter):
    identifier = "test_absolute_tokens"

    keywords = TestAdapter.keywords + (
        Keyword("optional", optional=True, help_text="A test keyword argument", type_hint="str", default="default"),
        Keyword("absolute", absolute=True, help_text="A test keyword argument", type_hint="bool"),
    )

    def render_content(self, parent_context: dict = None) -> str:
        return str(self.kwargs.get("absolute", False))

class TestBlockAdapter(BlockAdapter, TestAdapter):
    identifier = "test_block"

class TestFieldAdapter(FieldAdapter, TestAdapter):
    identifier = "test_field"

class TestModelAdapter(ModelAdapter, TestAdapter):
    identifier = "test_model"

adapter_registry.register(TestAdapter)
adapter_registry.register(TestBlockAdapter)
adapter_registry.register(TestFieldAdapter)
adapter_registry.register(TestModelAdapter)
adapter_registry.register(TestContextAdapter)
adapter_registry.register(TestAbsoluteTokensAdapter)

import uuid

def get_adapter_id() -> str:
    return str(uuid.uuid4())

class TestBaseAdapter(BaseFEditTest):

    def test_required_kwargs(self):
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))

    def test_absolute_tokens(self):
        self.assertEqual(TestAbsoluteTokensAdapter.absolute_tokens, tuple(["absolute"]))

    def test_required_kwargs_ok(self):
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))
        id = get_adapter_id()

        template_ok = Template(
            "{% load fedit %}"
            f"{{% fedit test object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        template_ok = template_ok.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        self.assertHTMLEqual(
            template_ok,
            f"TestAdapter: {self.basic_model.title} (test) ({id})"
        )

        self.assertDictEqual(
            adapters[id].kwargs,
            {"test": "test", "id": id}
        )

    def test_required_kwargs_fail(self):
        id = get_adapter_id()
        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        try:
            template_fail = Template(
                "{% load fedit %}"
                f"{{% fedit test object.title id='{id}' %}}"
            )

            # self.fail(f"Expected exception: {e}")
            template_fail.render(Context({
                "request": request,
                "object": self.basic_model,
            }))

        except TemplateSyntaxError as e:
            self.assertEqual(
                str(e),
                "Missing required keyword argument test"
            )

        except Exception as e:
            self.fail(f"Unexpected exception: {e} ({type(e)})")

        else:
            self.fail("Expected exception not raised")

    def test_optional_kwargs_default(self):
        id = get_adapter_id()
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_absolute_tokens object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        template = template.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        adapter = adapters[id]

        self.assertEqual(
            adapter.kwargs["optional"],
            "default",
        )

    def test_optional_kwargs_override(self):
        id = get_adapter_id()
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_absolute_tokens object.title optional='not default' test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        template = template.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        adapter = adapters[id]

        self.assertEqual(
            adapter.kwargs["optional"],
            "not default",
        )

    def test_adapter_absolute_tokens(self):
        id = get_adapter_id()
        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test_absolute_tokens object.title test='test' absolute id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        self.assertHTMLEqual(
            tpl,
            str(True),
        )

    def test_positional_kwargs(self):
        id = get_adapter_id()
        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test_absolute_tokens object.title 'test' absolute id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        adapter = adapters[id]

        self.assertEqual(
            adapter.kwargs["test"],
            "test",
        )

        self.assertEqual(
            adapter.kwargs["absolute"],
            True,
        )

    def test_adapter_absolute_tokens_fail(self):
        id = get_adapter_id()
        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test_absolute_tokens object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        self.assertHTMLEqual(
            tpl,
            str(False),
        )

    def test_adapter_render_content(self):
        adapter = TestAdapter(
            self.basic_model,
            "title",
            self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            ),
            id=3,
        )

        self.assertEqual(
            adapter.render_content(base_adapter_context(
                adapter=adapter,
                context={},
            )),
            f"TestAdapter: {self.basic_model.title} (3)"
        )

    def test_adapter_editable(self):
        id = get_adapter_id()
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))

        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], base_adapter_context(
                adapter=adapters[id],
                context={},
            ))
        )

    def test_adapter_editable_equals_refetch(self):
        id = get_adapter_id()
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))

        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        adapter = adapters[id]

        self.assertEqual(
            adapter.kwargs["test"],
            "test",
        )

        self.assertEqual(
            adapter.kwargs["id"],
            id,
        )

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], base_adapter_context(
                adapter=adapter,
                context={},
            ))
        )

        url = self.get_refetch_url(
            "test",
            self.basic_model._meta.app_label,
            self.basic_model._meta.model_name,
            self.basic_model.pk,
            "title",
        )

        self.client.force_login(self.admin_user)

        response = self.client.get(url, {
            "shared_context":\
                adapters[id].encode_shared_context(),
        })

        json_data = json.loads(
            response.content.decode("utf-8"),
        )

        if "html" not in json_data:
            self.fail(f"Expected 'html' in response, got {json_data}")

        if "success" not in json_data\
            or "success" in json_data and not json_data["success"]:
            self.fail(f"Expected 'success' in response, got {json_data}")

        if "refetch" not in json_data\
            or "refetch" in json_data and not json_data["refetch"]:
            self.fail(f"Expected 'refetch' in response, got {json_data}")

        self.assertHTMLEqual(
            tpl,
            json_data["html"],
        )


    def test_adapter_shared_context(self):
        uid = get_adapter_id()
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))

        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test object.title test='test' id='{uid}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        adapter = adapters[uid]

        self.assertEqual(
            adapter.kwargs["test"],
            "test",
        )

        self.assertEqual(
            adapter.kwargs["id"],
            uid,
        )

        new_id = get_adapter_id()
        adapter.kwargs["id"] = new_id
        context = adapter.encode_shared_context()
        adapter.kwargs["id"] = uid

        url = self.get_refetch_url(
            "test",
            self.basic_model._meta.app_label,
            self.basic_model._meta.model_name,
            self.basic_model.pk,
            "title",
        )

        self.client.force_login(self.admin_user)

        self.assertFalse(
            new_id in adapters,
        )

        response = self.client.get(url, {
            "shared_context": context,
        })

        # Load if lazy
        response.content.decode("utf-8")

        self.assertTrue(
            new_id in adapters,
        )

        self.assertFalse(
            id(adapter) == id(adapters[new_id]),
        )

        new_adapter = adapters[new_id]

        self.assertEqual(
            new_adapter.kwargs["test"],
            "test",
        )

        self.assertEqual(
            adapter.kwargs["test"],
            new_adapter.kwargs["test"],
        )

        self.assertEqual(
            new_adapter.kwargs["id"],
            new_id,
        )

        self.assertNotEqual(
            adapter.kwargs["id"],
            new_adapter.kwargs["id"],
        )

    def test_adapter_editable_as_var(self):
        id = get_adapter_id()
        self.assertEqual(TestAdapter.required_kwargs, tuple(["test"]))

        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test object.title test='test' id='{id}' as test_adapter_as_variable_name %}}"
            "{{ test_adapter_as_variable_name }}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
            })
        )

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], base_adapter_context(
                adapter=adapters[id],
                context={},
            ))
        )

    def test_context_processors_run(self):
        id = get_adapter_id()
        tpl = Template(
            "{% load fedit %}"
            f"{{% fedit test_context object.title test='test' id='{id}' %}}"
        )

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user

        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        # In this test we are testing the adapter - not the template.
        tpl = tpl.render(
            Context({
                "request": request,
                "object": self.basic_model,
                "testing": "testing context processor",
            }),
        )

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], base_adapter_context(
                adapter=adapters[id],
                context={},
            ), run_context_processors=True)
        )


class TestBlockAdapter(BaseFEditTest):

    def test_render(self):
        id = get_adapter_id()
        streamfield = self.basic_model.content
        block = find_block(self.BLOCK_ID, streamfield)
        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user
        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )
        block_value, _ = block
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_block object.content block=block block_id=block_id id='{id}' %}}"
        )

        context = {
            "object": self.basic_model,
            "request": request,
            "block": block_value,
            "block_id": self.BLOCK_ID,
        }

        tpl = template.render(Context(context))

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], {})
        )

    def test_render_as_var(self):
        id = get_adapter_id()
        streamfield = self.basic_model.content
        block = find_block(self.BLOCK_ID, streamfield)

        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )

        request.user = self.admin_user
        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )
        block_value, _ = block
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_block object.content block=block block_id=block_id id='{id}' as test %}}"
            "{{ test }}"
        )

        context = {
            "object": self.basic_model,
            "request": request,
            "block": block_value,
            "block_id": self.BLOCK_ID,
        }

        tpl = template.render(Context(context))

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], {})
        )

    def test_render_from_context(self):
        id = get_adapter_id()
        streamfield = self.basic_model.content
        block = find_block(self.BLOCK_ID, streamfield)
        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user
        setattr(
            request,
            FEDIT_PREVIEW_VAR,
            True,
        )
        block_value, _ = block
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_block from_context block=block block_id=block_id id='{id}' %}}"
        )

        context = {
            "object": self.basic_model,
            "request": request,
            "block": block_value,
            "block_id": self.BLOCK_ID,
            FIELD_TEMPLATE_VAR: "content",
            INSTANCE_TEMPLATE_VAR: self.basic_model,
        }

        tpl = template.render(Context(context))

        self.assertHTMLEqual(
            tpl,
            wrap_adapter(request, adapters[id], context)
        )


    def test_render_from_context_missing(self):
        id = get_adapter_id()
        streamfield = self.basic_model.content
        block = find_block(self.BLOCK_ID, streamfield)
        request = self.request_factory.get(
            self.get_editable_url(
                self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
            )
        )
        request.user = self.admin_user
        block_value, _ = block
        template = Template(
            "{% load fedit %}"
            f"{{% fedit test_block from_context block=block block_id=block_id id='{id}' %}}"
        )

        context = {
            "object": self.basic_model,
            "request": request,
            "block": block_value,
            "block_id": self.BLOCK_ID,
        }

        tpl = template.render(Context(context))

        self.assertHTMLEqual(
            tpl,
            block_value.render(context),
        )


class TestFieldAdapter(BaseFEditTest):
    
        def test_render(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_field object.title test=True id='{id}' %}}"
            )
    
            context = {
                "object": self.basic_model,
                "request": request,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], {})
            )

        def test_render_as_var(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_field object.title test=True id='{id}' as test %}}"
                "{{ test }}"
            )
    
            context = {
                "object": self.basic_model,
                "request": request,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], {})
            )

        def test_render_related_field(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_field object.related_field test=True id='{id}' as test %}}"
                "{{ test }}"
            )

            context = {
                "object": self.basic_model,
                "request": request,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], {})
            )

class TestModelAdapter(BaseFEditTest):
    
        def test_render(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_model object test=True id='{id}' %}}"
            )
    
            context = {
                "object": self.basic_model,
                "request": request,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], {})
            )
    
        def test_render_as_var(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_model object test=True id='{id}' as test %}}"
                "{{ test }}"
            )
    
            context = {
                "object": self.basic_model,
                "request": request,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], {})
            )
    
        def test_render_from_context(self):
            id = get_adapter_id()
            request = self.request_factory.get(
                self.get_editable_url(
                    self.basic_model.pk, self.basic_model._meta.app_label, self.basic_model._meta.model_name,
                )
            )
            request.user = self.admin_user
            setattr(
                request,
                FEDIT_PREVIEW_VAR,
                True,
            )
            template = Template(
                "{% load fedit %}"
                f"{{% fedit test_model from_context test=True id='{id}' %}}"
            )
    
            context = {
                "object": self.basic_model,
                "request": request,
                INSTANCE_TEMPLATE_VAR: self.basic_model,
            }
    
            tpl = template.render(Context(context))
    
            self.assertHTMLEqual(
                tpl,
                wrap_adapter(request, adapters[id], context)
            )



from ..adapters import (
    FieldAdapter,
    BlockAdapter,
    ModelAdapter,
    BackgroundImageFieldAdapter,
    adapter_registry,
)


adapter_registry.register(FieldAdapter)
adapter_registry.register(BlockAdapter)
adapter_registry.register(ModelAdapter)
adapter_registry.register(BackgroundImageFieldAdapter)




from typing import Type, Any
from django.template import (
    library, Node, TemplateSyntaxError,
)
from django.template.context import (
    Context,
)
from django.template.base import (
    Parser, Token,
    FilterExpression,
)
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.core import signing

from wagtail import hooks
from wagtail.models import (
    Page,
    PAGE_TEMPLATE_VAR,
)

import warnings

from ..settings import (
    TIPPY_ENABLED,
)
from ..adapters import (
    adapter_registry,
    RegistryLookUpError,
    BaseAdapter,
    AdapterError,
)
from ..utils import (
    wrap_adapter,
    with_userbar_model,
    base_adapter_context,
    _flatten_context,
    _can_edit,
    FEDIT_PREVIEW_VAR,
    TEMPLATE_TAG_NAME,

    FIELD_TEMPLATE_VAR,
    INSTANCE_TEMPLATE_VAR,
)
from ..hooks import (
    REGISTER_CSS,
    REGISTER_JS,
)


register = library.Library()


WARNING_FIELD_NAME_NOT_AVAILABLE = "Field name is not available in the context for %(object)s."
WARNING_MODEL_INSTANCE_NOT_AVAILABLE = "Model instance is not available in the context for %(object)s."


def as_var(var: str, context: Context, value: str) -> str:
    if not var:
        return value

    context[var] = value
    return ""

class AdapterNode(Node):
    signer = signing.TimestampSigner()

    def __init__(self, adapter: Type[BaseAdapter], model: FilterExpression, getters: list[str], as_var: str = None, **kwargs):
        self.adapter = adapter
        self.model = model
        self.getters = getters
        self.kwargs = kwargs
        self.as_var = as_var

    def _resolve_expressions(self, context, model, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, FilterExpression):
                kwargs[k] = v.resolve(context)
        
        if isinstance(model, FilterExpression):
            model = model.resolve(context)

        return model, kwargs

    def render(self, context):
        model = self.model
        getters = self.getters

        model, kwargs = self._resolve_expressions(
            context, model, **self.kwargs,
        )

        if FIELD_TEMPLATE_VAR in context\
            and INSTANCE_TEMPLATE_VAR in context\
            and not model and self.adapter.field_required:

            field_name = context[FIELD_TEMPLATE_VAR]
            obj = context[INSTANCE_TEMPLATE_VAR]

        elif not model\
                and INSTANCE_TEMPLATE_VAR in context\
                and not self.adapter.field_required:
            obj = context[INSTANCE_TEMPLATE_VAR]
            field_name = None
            
        else:

            if not model:
                warnings.warn(
                    WARNING_MODEL_INSTANCE_NOT_AVAILABLE % {
                        "object": self.adapter.__name__,
                    },
                    RuntimeWarning,
                )

                context = context.flatten()

                try:
                    return as_var(self.as_var, context, self.adapter.render_from_kwargs(
                        context, **kwargs,
                    ))
                except AdapterError as e:
                    raise TemplateSyntaxError(str(e))

            field_name = None
            obj = model
            if getters:
                field_name = getters[len(getters) - 1]

                for i in range(len(getters) - 1):
                    getter = getters[i]
                    try:
                        obj = getattr(obj, getter)
                    except AttributeError:
                        raise TemplateSyntaxError(f"Object {model.__class__.__name__} does not have attribute {getter}")
                    

        request = context.get("request")
        adapter = self.adapter(
            object=obj,
            field_name=field_name,
            request=request,
            **kwargs,
        )

        context = base_adapter_context(
            adapter,
            context,
        )

        content = None
        if adapter.check_permissions()\
          and _can_edit(request, obj):
            content = wrap_adapter(
                request=request,
                adapter=adapter,
                context=context,
                run_context_processors=False,
            )

        else:
            _flatten_context(context)
            content = adapter.render_content(
                context,
            )

        return as_var(
            self.as_var,
            context,
            content,
        )


@register.tag(name=TEMPLATE_TAG_NAME)
def do_render_fedit(parser: Parser, token: Token):

    tokens = token.split_contents()

    _ = tokens.pop(0)
    adapter_id = tokens.pop(0)

    try:
        adapter = adapter_registry[adapter_id]
    except RegistryLookUpError:
        raise TemplateSyntaxError(f"No adapter found with identifier '{adapter_id}'")

    model, model_tokens = None, None
    if tokens:
        model__field = tokens.pop(0)
        model_tokens = model__field.split(".")
        
        # If the field is required; the length must be at least 2,
        # or the first token must be "from_context"
        if len(model_tokens) < 2:
            if model_tokens[0] != "from_context" and adapter.field_required:
                raise TemplateSyntaxError(
                    "Model and field name are required: 'mymodel.myfield' or 'from_context'",
                )

        # If the field is not required; the length could be 1 by specifying only a model.
        if len(model_tokens) > 1 or (
            model_tokens[0] != "from_context" and not adapter.field_required
        ):
            # mymodel
            # mymodel.myfield
            # mymodel.related_field.myfield
            model = parser.compile_filter(model_tokens.pop(0))

    as_var = None
    if tokens and tokens[len(tokens)-2] == "as":
        as_var = tokens.pop(len(tokens)-1)
        tokens.pop(len(tokens)-1)

    kwargs = get_kwargs(
        parser,
        tokens,
        adapter.required_kwargs,
        adapter.absolute_tokens,
    )

    return AdapterNode(
        adapter=adapter,
        model=model,
        getters=model_tokens,
        as_var=as_var,
        **kwargs,
    )


@register.simple_tag(takes_context=True)
def render_adapter(context: Context, adapter: BaseAdapter) -> str:
    adapter_context = {}
    
    if "adapter_context" in context:
        adapter_context = context["adapter_context"]
        del context["adapter_context"]

    context = _flatten_context(
        context,
    )
    adapter_context = _flatten_context(
        adapter_context,
    )

    context.update(adapter_context)

    return adapter.render_content(
        context,
    )


@register.inclusion_tag("wagtail_fedit/_hook_output.html", name="fedit_scripts", takes_context=True)
def static_hook_output(context, css_or_js) -> dict:
    css_or_js = css_or_js.lower()

    if css_or_js not in ["css", "js"]:
        raise ValueError("Invalid argument, must be 'css' or 'js'")

    request = context.get("request")
    
    if css_or_js == "css":
        hook_name = REGISTER_CSS
    else:
        hook_name = REGISTER_JS

    if not getattr(request, FEDIT_PREVIEW_VAR, False):
        return {}

    files = []
    for hook in hooks.get_hooks(hook_name):
        ret = hook(request)
        
        if not isinstance(ret, (list, tuple)):
            ret = [ret]

        files.extend(ret)
        
    return {
        "hook_output": files,
    }

@register.simple_tag(takes_context=False, name="tooltip")
def tooltip(content, wrapping: str = None, **kwargs) -> str:

    if not TIPPY_ENABLED:
        return ""

    kwargs["content"] = content
    s = [
        "data-tooltip='true'"
    ]
    for key, value in kwargs.items():
        s.append(f"data-tooltip-{key}='{escape(value)}'")
    attrs = " ".join(s)

    if not wrapping:
        return mark_safe(attrs)
    
    return mark_safe(f"<span {attrs}>{wrapping}</span>")

@register.simple_tag(takes_context=True, name="fedit_userbar")
def do_with_userbar_model(context, model: Any) -> str:
    if "request" in context:
        context["request"] = with_userbar_model(context["request"], model)
    if PAGE_TEMPLATE_VAR in context and context[PAGE_TEMPLATE_VAR] != model:
        context[PAGE_TEMPLATE_VAR] = model
    return ""

def get_kwargs(parser: Parser, tokens: list[str], kwarg_list: list[str] = None, absolute_tokens: list[str] = None) -> dict:
    had_kwargs = False
    kwargs = {}

    if not kwarg_list:
        kwarg_list = tuple()

    if not absolute_tokens:
        absolute_tokens = tuple()

    for i, token in enumerate(tokens):
        split = token.split("=")
        if len(split) == 1 and len(kwarg_list) > i:
            if split[0] in absolute_tokens:
                kwargs[split[0]] = True
                continue

            if had_kwargs:
                raise ValueError("Unexpected positional argument after keyword argument")
            
            kwargs[kwarg_list[i]] = parser.compile_filter(token)
        elif len(split) == 1:
            if split[0] in absolute_tokens:
                kwargs[split[0]] = True
                continue
            else:
                key = split[0]
                value = parser.compile_filter(key)
                kwargs[key] = value
        else:
            key = split[0]
            # if key not in kwargs_names:
            #     raise ValueError(f"Unexpected keyword argument {key}")
            if key in absolute_tokens:
                raise TemplateSyntaxError(
                    f"Keyword argument {key} cannot be resolved;"
                    " it can only be used as an absolute argument."
                )
            
            kwargs[key] = parser.compile_filter(split[1])
            had_kwargs = True

    for key in kwarg_list:
        if key not in kwargs:
            raise TemplateSyntaxError(
                f"Missing required keyword argument {key}"
            )

    return kwargs




from django.forms.models import model_to_dict
from wagtail.admin.panels import (
    FieldPanel,
    ObjectList,
)
from wagtail.models import (
    RevisionMixin,
)
from .base import (
    BaseFEditTest,
)

import json

class TestModelEdit(BaseFEditTest):
    def test_model_edited(self):
        self.client.force_login(self.admin_user)
        
        for i, model in enumerate(self.models):
            
            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 0)
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()

            panels = [
                FieldPanel("title"),
                FieldPanel("body"),
            ]

            setattr(model.__class__, "panels", panels)
            setattr(model.__class__, "edit_handler", ObjectList(panels))

            initial_title = model.title
            d = model_to_dict(model, fields=["title", "body"])
            d["title"] = f"{model.title} test case {i + 1}"

            self.client.post(
                self.get_model_url(
                    model._meta.app_label,
                    model._meta.model_name,
                    model.pk,
                ),
                d,
            )

            model.refresh_from_db()

            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 1, msg=f"Expected 1 revision, got {model.revisions.count()} for {model.__class__.__name__} '{model}'")
                self.assertEqual(model.title, initial_title, msg=f"Expected title to be '{initial_title}', got '{model.title}' for {model.__class__.__name__} '{model}'")
                chk = model.latest_revision.as_object()
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()
                chk = model

            chk_title = f"{initial_title} test case {i + 1}"
            self.assertEqual(chk.title, chk_title, msg=f"Expected title to be '{chk_title}', got '{chk.title}' for {model.__class__.__name__} '{model}'")

    def test_unauthorized_unchanged(self):
        self.client.force_login(self.regular_user)
        
        for i, model in enumerate(self.models):
            
            initial_title = model.title
            model.__class__.panels = [
                FieldPanel("title"),
                FieldPanel("body"),
            ]
            model.__class__.edit_handler = ObjectList(model.__class__.panels)
            d = model_to_dict(model, fields=["title", "body"])
            d["title"] = f"{model.title} test case {i + 1}"
            response = self.client.post(
                self.get_model_url(
                    model._meta.app_label,
                    model._meta.model_name,
                    model.pk,
                ),
                d,
            )

            self.assertEqual(response.status_code, 403, msg=f"Expected 403, got {response.status_code} ({response.content})")

            model.refresh_from_db()

            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 0)
                chk = model

            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()
                chk = model

            self.assertEqual(chk.title, initial_title)


    def test_lock_unchanged(self):
        self.client.force_login(self.other_admin_user)
        initial_title = self.lock_model.title

        panels = self.lock_model.__class__.panels
        self.lock_model.__class__.panels = [
            FieldPanel("title"),
            FieldPanel("body"),
        ]

        d = model_to_dict(self.lock_model, fields=["title", "body"])
        d["title"] = f"{self.lock_model.title} test case"

        response = self.client.post(
            self.get_model_url(
                self.lock_model._meta.app_label,
                self.lock_model._meta.model_name,
                self.lock_model.pk,
            ),
        )

        self.lock_model.__class__.panels = panels
        self.lock_model.refresh_from_db()

        self.assertEqual(response.status_code, 423, msg=str(response.content)) # 423 Locked
        self.assertEqual(self.lock_model.revisions.count(), 0)
        
        try:
            response_content = (json.loads(response.content) or {})
        except json.JSONDecodeError:
            self.fail("Response content is not valid JSON")
            
        self.assertTrue(response_content.get("locked", False))
        self.assertEqual(self.lock_model.locked_by, self.admin_user)
        self.assertEqual(self.lock_model.title, initial_title)

        
        




from wagtail.models import (
    RevisionMixin,
)
from wagtail_fedit.utils import (
    find_block,
)
from .base import (
    BaseFEditTest,
)

import json

class TestBlockEdit(BaseFEditTest):
    def test_block_edited(self):

        self.client.force_login(self.admin_user)
        
        for i, model in enumerate(self.models):
            
            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 0)
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()

            bound, _ = find_block(block_id=self.BLOCK_ID, field=model.content)
            initial_content = bound.value["link"]["text"]

            response = self.client.post(
                self.get_block_url(
                    self.BLOCK_ID,
                    "content",
                    model._meta.app_label,
                    model._meta.model_name,
                    model.pk,
                ),
                {
                    "value-link-text": f"{initial_content} test case {i + 1}"
                }
            )

            self.assertEqual(response.status_code, 200)
            model.refresh_from_db()

            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 1)
                chk = model.latest_revision.as_object()
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()
                chk = model

            bound, contentpath = find_block(block_id=self.BLOCK_ID, field=chk.content)
            self.assertEqual(bound.value["link"]["text"], f"{initial_content} test case {i + 1}",
                msg=f"Block: {bound.block} does not contain the expected value: {initial_content} test case {i + 1}"
            )

    def test_unauthorized_unchanged(self):
        self.client.force_login(self.regular_user)
        
        for i, model in enumerate(self.models):
            
            initial_content = model.content
            response = self.client.post(
                self.get_field_url(
                    "content",
                    model._meta.app_label,
                    model._meta.model_name,
                    model.pk,
                ),
                {
                    "content": f"{model.content} test case {i + 1}"
                }
            )

            self.assertEqual(response.status_code, 403, msg=f"Expected 403, got {response.status_code} ({response.content})")

            model.refresh_from_db()

            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 0)
                chk = model

            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()
                chk = model

            self.assertEqual(chk.content.get_prep_value(), initial_content.get_prep_value())

    def test_lock_unchanged(self):
        self.client.force_login(self.other_admin_user)
        initial_content = self.lock_model.content
        initial_bound, _ = find_block(block_id=self.BLOCK_ID, field=self.lock_model.content)

        response = self.client.post(
            self.get_block_url(
                self.BLOCK_ID,
                "content",
                self.lock_model._meta.app_label,
                self.lock_model._meta.model_name,
                self.lock_model.pk,
            ),
            {
                "value-link-text": f"{initial_bound.value['link']['text']} test case"
            }
        )

        self.lock_model.refresh_from_db()

        self.assertEqual(response.status_code, 423) # 423 Locked
        self.assertEqual(self.lock_model.revisions.count(), 0)
        
        try:
            response_content = (json.loads(response.content) or {})
        except json.JSONDecodeError:
            self.fail("Response content is not valid JSON")
            
        self.assertTrue(response_content.get("locked", False))
        self.assertEqual(self.lock_model.locked_by, self.admin_user)
        self.assertEqual(self.lock_model.content.get_prep_value(), initial_content.get_prep_value())



from wagtail import hooks


@hooks.register("register_icons")
def register_icons(icons):
    icons.extend([
        "wagtail_fedit/icons/fedit-save.svg",
        "wagtail_fedit/icons/fedit-check-list.svg",
        "wagtail_fedit/icons/fedit-eye-open.svg",
        "wagtail_fedit/icons/fedit-eye-closed.svg",
        "wagtail_fedit/icons/fedit-stop-sign.svg",
    ])

    return icons



from django.utils.translation import (
    gettext_lazy as _,
    gettext,
)
from wagtail.log_actions import LogFormatter
from wagtail import hooks

@hooks.register("register_log_actions")
def register_core_log_actions(actions):
    actions.register_action("wagtail_fedit.related_changed", _("Edit Related"), _("Related Object Edited"))

    @actions.register_action("wagtail_fedit.edit_field")
    class FieldChangedFormatter(LogFormatter):
        label = _("Field Changed (Frontend)")
        message = _("A field was changed from the frontend")

        def format_message(self, log_entry):
            data = log_entry.data

            if not "verbose_field_name" in data:
                return self.label

            if "edited_model_string" in data\
                and "edited_model_verbose" in data:
                return _("Changed related '%(related)s' instance '%(instance)s' on field '%(field)s' (Frontend)") % {
                    "field": gettext(data["verbose_field_name"]),
                    "related": gettext(data["edited_model_verbose"]),
                    "instance": data["edited_model_string"],
                }

            try:

                old = data["old"]
                new = data["new"]

                if len(old) > 50:
                    old = old[:50] + "..."

                if len(new) > 50:
                    new = new[:50] + "..."

                return _("Changed '%(field)s' from '%(old)s' to '%(new)s' (Frontend)") % {
                    "field": gettext(data["verbose_field_name"]),
                    "old": old,
                    "new": new,
                }
            
            except KeyError:
                return _("Edit Field")

    @actions.register_action("wagtail_fedit.edit_model")
    class ModelChangedFormatter(LogFormatter):
        label = _("Model Changed (Frontend)")
        message = _("A model was changed from the frontend")

        def format_message(self, log_entry):
            data = log_entry.data

            if not "model_verbose" in data:
                return self.label
            
            return _("Changed '%(model)s' (Frontend)") % {
                "model": gettext(data["model_verbose"]),
            }
            
    @actions.register_action("wagtail_fedit.edit_block")
    class BlockChangedFormatter(LogFormatter):
        label = _("Block Changed (Frontend)")
        message = _("A block was changed from the frontend")
        must = [
            "block_id",
            "verbose_field_name",
            "block_label",
        ]

        def format_message(self, log_entry):
            data = log_entry.data

            if not all([key in data for key in self.must]):
                return self.label

            return _("Changed block \"%(block)s\" on field \"%(field)s\" (%(block_id)s, Frontend)") % {
                "block": gettext(data["block_label"]),
                "field": gettext(data["verbose_field_name"]),
                "block_id": data["block_id"],
            }


from django.test import TestCase
from wagtail import blocks
from wagtail_fedit import forms as block_forms
from wagtail_fedit import utils
from .base import TEST_BLOCK_DATA
from ..models import (
    HeadingComponent,
    FlatMenuComponent
)

import uuid

_model_map = {}

class FakeModel(object):
    def __init__(self, title="Fake Model"):
        self.pk = uuid.uuid4()
        self.title = title
        self.attrs = {}

    def __setattr__(self, key, value):
        if key in ["pk", "title", "attrs"]:
            object.__setattr__(self, key, value)
        else:
            self.attrs[key] = value

    def __getattribute__(self, key):
        if key in ["pk", "title", "attrs", "full_clean", "save"]:
            return object.__getattribute__(self, key)
        return self.attrs.get(key)
    
    def full_clean(self):
        pass

    def save(self):
        _model_map[self.pk] = self

class FakeUser:
    def __init__(self, is_authenticated=True, is_staff=True):
        self.is_authenticated = is_authenticated
        self.is_staff = is_staff

class FakeRequest:
    def __init__(self, method="GET", user=None, GET=None, POST=None):
        self.user = user or FakeUser()
        self.method = method
        self.GET = GET
        self.POST = POST

class TestBlocks(TestCase):
    def setUp(self) -> None:
        self.stream_block = blocks.StreamBlock([
            ("heading_component", HeadingComponent()),
            ("flat_menu_component", FlatMenuComponent())
        ])
        self.stream_value = self.stream_block.to_python(TEST_BLOCK_DATA)

    def test_find_block(self):
        block, contentpath = utils.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)
        self.assertEqual(block.value["heading"], "AWESOME!!")
        self.assertEqual(block.value["subheading"], "RIGHT?!")

        block, contentpath = utils.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)
        self.assertEqual(block.value["link"]["text"], "Test Item 3")
        self.assertEqual(contentpath, ["3e9144fd-5fa5-47f8-917e-8fe87c15da01", "items", "a98a19c6-2ead-4e69-9ea2-3158c7e82976"])

        item = self.stream_block.get_block_by_content_path(self.stream_value, contentpath)
        self.assertEqual(item.value["link"]["text"], "Test Item 3")

        block, contentpath = utils.find_block("invalid-id", self.stream_value)
        self.assertIsNone(block)
        self.assertEqual(contentpath, [])

    def test_get_form_class(self):
        block = utils.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)[0]
        form_class = block_forms.get_block_form_class(block.block)
        self.assertIsNotNone(form_class)

        VALID_DATA = {
            "value-heading": "New Heading",
            "value-subheading": "New Subheading"
        }

        INVALID_DATA = {
            "value-heading": "New Heading",
            "value-subheading": "New Subheading" * 10
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel())
        self.assertTrue(valid_form.is_valid())

        invalid_form = form_class(data=INVALID_DATA, block=block, parent_instance=FakeModel())
        self.assertFalse(invalid_form.is_valid())

    def test_get_subblock_form_class(self):
        block = utils.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)[0]
        form_class = block_forms.get_block_form_class(block.block)
        self.assertIsNotNone(form_class)

        VALID_DATA = {
            "value-link-text": "New Link Text"
        }

        INVALID_DATA = {
            "value-link-text": ["New Link Text" * 10]
        }


        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel())
        if not valid_form.is_valid():
            self.fail((
                valid_form.errors.items(),
                valid_form.non_field_errors()
            ))

        valid_form.save()
        self.assertEqual(block.value["link"]["text"], "New Link Text")

        invalid_form = form_class(data=INVALID_DATA, block=block, parent_instance=FakeModel())
        self.assertFalse(invalid_form.is_valid())


    def test_block_form_save(self):
        block = utils.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)[0]
        form_class = block_forms.get_block_form_class(block.block)

        VALID_DATA = {
            "value-heading": "New Heading",
            "value-subheading": "New Subheading"
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel())
        if not valid_form.is_valid():
            self.fail((valid_form.errors.items(), valid_form.non_field_errors()))

        valid_form.save()
        self.assertEqual(block.value["heading"], "New Heading")
        self.assertEqual(block.value["subheading"], "New Subheading")


    def test_subblock_form_save(self):
        block = utils.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)[0]
        form_class = block_forms.get_block_form_class(block.block)

        VALID_DATA = {
            "value-link-text": "New Link Text"
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel())
        if not valid_form.is_valid():
            errors = valid_form.errors.as_data()["value"][0].as_json_data()
            self.fail((errors, valid_form.non_field_errors()))

        valid_form.save()
        self.assertEqual(block.value["link"]["text"], "New Link Text")



from django.utils.translation import gettext_lazy as _

class Formatter:
    def __init__(self, *messages):
        self.messages = messages

    def format(self, *args, **kwargs):
        messages = list(self.messages)
        if kwargs:
            raise ValueError("Formatter does not support keyword arguments")
        
        if len(args) > len(messages):
            raise ValueError("Number of arguments does not match number of messages")
        
        made = []
        
        for i, arg in enumerate(args):
            made.append(messages[i].format(arg))

        return " ".join(made)

NO_PERMISSION_VIEW = _("You do not have permission to view this page.")
NO_PERMISSION_ACTION = _("User does not have permission to {}")
NO_WORKFLOW_STATE = _("No workflow state found")
OBJECT_LOCKED = _("This object is locked - it cannot be acted upon.")
OBJECT_NOT_LIVE = _("This object is not live")
ACTION_MISSING = _("No action specified")
INVALID_ACTION = _("Invalid action specified: {}")
MISSING_REQUIRED_SUPERCLASSES = _("Model {} does not inherit from {}")
NO_UNPUBLISHED_CHANGES = _("Object has no unpublished changes")
MODEL_NOT_FOUND = _("Model not found")
INVALID = Formatter(
    _("Invalid {}"),
    _("for object {}"),
)
REQUIRED = Formatter(
    _("{} is required"),
    _("for object {}"),
)



from django.urls import path
from . import views

app_name = "wagtail_fedit"

urlpatterns = []

adapter_based_views = (
    ("edit", views.EditAdapterView),
    ("refetch", views.AdapterRefetchView),
)

for name, view in adapter_based_views:
    view.url_name = f"wagtail_fedit:{name}"
    view.url_pattern = f"{name}/<str:adapter_id>/<str:app_label>/<str:model_name>/<str:model_id>/<str:field_name>/"
    urlpatterns.append(
        path(view.url_pattern, view.as_view(), name=name)
    )
    view.url_pattern = f"{name}/<str:adapter_id>/<str:app_label>/<str:model_name>/<str:model_id>/"
    urlpatterns.append(
        path(view.url_pattern, view.as_view(), name=name)
    )

# Model based views
model_based_views = (
    ("editable", views.FEditableView),
    ("publish", views.PublishView),
    ("submit", views.SubmitView),
    ("unpublish", views.UnpublishView),
    ("cancel", views.CancelView),
)

for name, view in model_based_views:
    view.url_name = f"wagtail_fedit:{name}"
    view.url_pattern = f"{name}/<str:object_id>/<str:app_label>/<str:model_name>/"
    urlpatterns.append(
        path(view.url_pattern, view.as_view(), name=name)
    )



from typing import Type
from .base import BaseAdapter


class RegistryLookUpError(Exception):
    pass

class DuplicateAdapterError(Exception):
    pass

class AdapterSubclassError(TypeError):
    pass


class AdapterRegistry:
    def __init__(self):
        self.adapters: dict[str, Type[BaseAdapter]] = {}

    def __getitem__(self, identifier: str) -> Type[BaseAdapter]:
        """
        Retrieve an adapter by its identifier or raise a RegistryLookUpError if not found.
        """
        try:
            return self.adapters[identifier]
        except KeyError:
            raise RegistryLookUpError(f"No adapter found with identifier '{identifier}'.")

    def register(self, adapter_class: Type[BaseAdapter]):
        """
        Register an adapter class with the registry.
        """
        if isinstance(adapter_class, BaseAdapter):
            raise AdapterSubclassError(f"{adapter_class.__class__.__name__} must be a subclass of BaseAdapter; got instance.")
        
        if adapter_class.identifier in self.adapters:
            raise DuplicateAdapterError(f"An adapter with identifier '{adapter_class.identifier}' is already registered.")

        self.adapters[adapter_class.identifier] = adapter_class

    def unregister(self, identifier):
        """
        Unregister an adapter by its identifier.
        """
        if identifier not in self.adapters:
            raise RegistryLookUpError(f"No adapter found with identifier '{identifier}'.")

        del self.adapters[identifier]


registry = AdapterRegistry()



from typing import Any
from django.apps import apps
from django.utils.translation import gettext as _
from django.utils.decorators import method_decorator
from django.template.loader import render_to_string
from django.views.decorators.clickjacking import (
    xframe_options_sameorigin,
)
from django.views.generic import View
from django.shortcuts import render
from django.http import (
    HttpRequest,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    JsonResponse,
    HttpResponse,
)
from wagtail.models import (
    RevisionMixin,
    PAGE_TEMPLATE_VAR,
    Page,
)
from wagtail.admin.views.generic import (
    WagtailAdminTemplateMixin,
)

from ..adapters import (
    adapter_registry,
    BaseAdapter,
    RegistryLookUpError,
)
from ..utils import (
    FeditPermissionCheck,
    FeditIFrameMixin,
    FEDIT_PREVIEW_VAR,
    base_adapter_context,
    lock_info,
)
from ..errors import (
    NO_PERMISSION_ACTION,
    INVALID,
    REQUIRED,
)
from .mixins import (
    LocaleMixin,
)

@method_decorator(xframe_options_sameorigin, name="dispatch")
class BaseAdapterView(FeditIFrameMixin, FeditPermissionCheck, WagtailAdminTemplateMixin, View):
    ERROR_TITLE = _("Validation Errors")

    def dispatch(self, 
            request:    HttpRequest,
            adapter_id: str = None,
            app_label:  str = None,
            model_name: str = None,
            model_id:   Any = None,
            field_name: str = None,
        ) -> None:

        # Fetch the adapter class from the registry
        try:
            self.adapter_class: BaseAdapter = adapter_registry[adapter_id]
        except RegistryLookUpError:
            return HttpResponseBadRequest(
                INVALID.format(
                    _("Adapter ID")
                )
            )

        # Retrieve the model class
        try:
            self.model = apps.get_model(app_label, model_name)
        except LookupError:
            return HttpResponseBadRequest(
                INVALID.format(
                    _("Model")
                )
            )
        
        # Check if the user has permissions to view the page
        if not self.has_perms(request, self.model):
            return HttpResponseForbidden(NO_PERMISSION_ACTION.format(
                _("view this page")
            ))

        # Only fetch latest reivision if it exists
        # If not; it will be automatically created by the form.
        model_instance = self.model._default_manager.get(pk=model_id)
        if isinstance(model_instance, RevisionMixin) and model_instance.latest_revision_id:
            self.instance = model_instance.latest_revision.as_object()
        else:
            self.instance = model_instance

        LocaleMixin.setup_locale(
            self.instance,
        )

        if not field_name and self.adapter_class.field_required:
            return HttpResponseBadRequest(
                REQUIRED.format(
                    _("Field name"),
                    self.instance,
                )
            )

        if field_name and not hasattr(self.instance, field_name) and self.adapter_class.field_required:
            return HttpResponseBadRequest(
                INVALID.format(
                    _("field name"),
                    self.instance,
                )
            )

        shared_context = request.GET.get("shared_context")
        if shared_context:
            self.shared_context = self.adapter_class.decode_shared_context(
                request,
                self.instance,
                field_name,
                shared_context,
            )
        else:
            self.shared_context = {}

        self.adapter = self.adapter_class(
            request=request,
            object=self.instance,
            field_name=field_name,
            **self.shared_context,
        )

        if not self.adapter.check_permissions():
            return HttpResponseForbidden(
                NO_PERMISSION_ACTION.format(
                    _("edit this field")
                )
            )

        self.lock, self.locked_for_user = lock_info(
            self.adapter.object, request.user,
        )

        setattr(
            self.request,
            FEDIT_PREVIEW_VAR,
            True,
        )

        return super().dispatch(
            request,
            adapter_id=adapter_id,
            field_name=field_name,
            app_label=app_label,
            model_name=model_name,
            model_id=model_id,
        )
    
    @property
    def template_name(self):
        return self.adapter.get_template_names()
    
    def render_to_response(self, context: dict[str, Any], success: bool = True, extra: dict = None, **response_kwargs: Any) -> HttpResponse:
        if not extra:
            extra = {}

        extra.update({
            "success": success,
        })

        context.update(extra)

        return render(
            self.request,
            self.template_name,
            context,
        )
            
    def get_header_title(self):
        return self.adapter.get_header_title()

    def get_help_text(self):
        return self.adapter.get_help_text()
    
    def get_context_data(self, **kwargs):
        shared_context = None
        if self.shared_context:
            shared_context =\
                self.request.GET["shared_context"]

        verbose_name = self.model._meta.verbose_name
        if self.adapter.field_required:
            verbose_name = self.adapter.meta_field.verbose_name

        extra = {}

        if isinstance(self.instance, Page):
            # Add the page template variable to the context.
            # Wagtail uses this internally; for example in `{% wagtailpagecache %}`
            extra[PAGE_TEMPLATE_VAR] = self.instance

        # Form context; used for rendering the modal.
        if "form" in kwargs:
            extra.update({
                "verbose_name": verbose_name,
                "locked_for_user": self.locked_for_user,
                "shared_context": shared_context,
                "form_attrs": self.adapter.get_form_attrs(),
                "locked": self.lock is not None,
                **self.adapter.get_form_context(
                    **kwargs,
                ),
            })
        # Add adapter context instead of form context
        else:
            extra = base_adapter_context(
                self.adapter,
                extra,
            )

        return super().get_context_data(**kwargs)\
            | extra
    
class AdapterRefetchView(BaseAdapterView):
    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        context = self.get_context_data()

        return JsonResponse({
            "success": True,
            "refetch": True,
            **self.adapter\
              .get_response_data(
                  context,
              ),
        })

class EditAdapterView(BaseAdapterView):
    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        # Can omit data from context - we are not rendering the content.
        form = self.adapter.get_form()

        return self.render_to_response(
            self.get_context_data(
                form=form,
            ),
            success=True,
        )

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        form = self.adapter.get_form()

        if not form.is_valid() or self.locked_for_user:

            self.adapter.form_invalid(form)

            return JsonResponse({
                "success": False,
                "errors": form.errors,
                "locked": self.locked_for_user,
                "html": render_to_string(
                    self.template_name,
                    context=self.get_context_data(
                        form=form,
                    ),
                    request=request,
                )
            }, status=423 if self.locked_for_user else 400)

        
        self.adapter.form_valid(form)

        context = self.get_context_data()

        return JsonResponse({
            "success": True,
            **self.adapter\
              .get_response_data(context),
        })




from wagtail.models import (
    RevisionMixin,
)
from ..models import (
    EditableFullModel,
    EditableDraftModel,
    EditableRevisionModel,
    EditablePreviewModel,
)

from .base import (
    BaseFEditTest,
)

class TestRevisions(BaseFEditTest):
    def test_revision_creation(self):
        self.client.force_login(self.admin_user)
        
        for i, model in enumerate(self.models):
            
            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 0)
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()

            initial_title = model.title

            self.client.post(
                self.get_field_url(
                    "title",
                    model._meta.app_label,
                    model._meta.model_name,
                    model.pk,
                ),
                {
                    "title": f"{model.title} test case {i + 1}"
                }
            )

            model.refresh_from_db()

            if isinstance(model, RevisionMixin):
                self.assertEqual(model.revisions.count(), 1)
                self.assertEqual(model.title, initial_title)
                chk = model.latest_revision.as_object()
            else:
                with self.assertRaises(AttributeError):
                    model.revisions.count()
                chk = model

            self.assertEqual(chk.title, f"{initial_title} test case {i + 1}")






from django.utils.html import format_html
from django.templatetags.static import static
from wagtail import hooks
from wagtail.models import Page
from wagtail.admin.widgets import (
    AdminPageChooser,
)
from wagtail.fields import (
    RichTextField,
    StreamField,
)
from wagtail.templatetags.wagtailcore_tags import (
    richtext,
)
from wagtail.images import (
    get_image_model,
)
from wagtail.images.widgets import AdminImageChooser
from wagtail.documents import get_document_model
from wagtail.documents.widgets import AdminDocumentChooser
from ..settings import (
    TIPPY_ENABLED,
)
from ..hooks import (
    EXCLUDE_FROM_RELATED_FORMS,
    REGISTER_FIELD_RENDERER,
    REGISTER_FIELD_WIDGETS,
    FIELD_EDITOR_SIZE,
    REGISTER_CSS,
    REGISTER_JS,
)

Image = get_image_model()
Document = get_document_model()

@hooks.register(EXCLUDE_FROM_RELATED_FORMS)
def exclude_related_forms(field):
    if field.related_model in [Page, Image, Document]:
        return True
    return False


@hooks.register(REGISTER_FIELD_RENDERER)
def register_renderers(renderer_map):
    # This is a custom renderer for RichText fields.
    # It will render the RichText field as a RichText block.
    renderer_map[RichTextField] = lambda request, context, instance, value: richtext(value)

@hooks.register(REGISTER_FIELD_WIDGETS)
def register_field_widgets(widgets):
    widgets[Image] = AdminImageChooser
    widgets[Document] = AdminDocumentChooser
    widgets[Page] = AdminPageChooser
    return widgets

@hooks.register(FIELD_EDITOR_SIZE)
def field_editor_size(model_instance, model_field):
    if isinstance(model_field, RichTextField):
        return "large"
    
    if isinstance(model_field, StreamField):
        return "full"
    
    return None


# <link rel="stylesheet" href="{% static 'wagtail_fedit/css/frontend.css' %}">
# <script src="{% static 'wagtail_fedit/js/frontend.js' %}"></script>

@hooks.register(REGISTER_CSS, order=-1)
def register_css(request):
    return [
        format_html(
            '<link rel="stylesheet" href="{0}">',
            static('wagtail_fedit/css/frontend.css')
        ),
    ]

@hooks.register(REGISTER_JS, order=-1)
def register_js(request):
    scripts = [
        format_html(
            '<script src="{0}"></script>',
            static('wagtail_fedit/js/edit.js')
        ),
    ]
    return scripts
